#ifndef DAMIERDYNG_H
#define DAMIERDYNG_H

#include <iostream>
using namespace std;
#include <cmath>
#include <cstdlib>

template <class G>
class DamierDynG
{
public:
    DamierDynG(int l, int c, G vd=0);
    DamierDynG(const DamierDynG &D);
    ~DamierDynG();

    DamierDynG<G>& operator=  (const DamierDynG<G> &D); // opérateur d'affectation
    DamierDynG<G>& operator+= (const DamierDynG<G> &D);
    DamierDynG<G>& operator+= (G c);
    DamierDynG<G>  operator+  (const DamierDynG<G> &D);
    template <class B> friend ostream& operator<< (ostream& sortie, DamierDynG<B>& V);

    void Print();
    void Init(G value);
    void Set(int x, int y, G value);
    void ReDim(int l, int c, G vd = 0);

private:
    int L;
    int C;
    G** T;

    // Méthode privée (factorisation  de code)
    void Alloc(int l, int c);
    void Free();
    bool sameDimensions (const DamierDynG<G> &D);
};

template<class G>
DamierDynG<G>::DamierDynG(int l, int c, G vd)
{
    Alloc(l, c);
    Init(vd);
}

template<class G>
DamierDynG<G>::DamierDynG(const DamierDynG<G> &D)
{
    Alloc(D.L, D.C);
    for(int i=0; i<L; i++)
        for(int j=0; j<C; j++)
            T[i][j] = D.T[i][j];
}

template<class G>
DamierDynG<G>::~DamierDynG(){
    if (T != NULL) {
        Free();
        T = NULL;
    }
}

template<class G>
void DamierDynG<G>::Free(){
    for (int i=0; i<L; i++)
        delete [] T[i];
    delete [] T;
}

template<class G>
void DamierDynG<G>::Alloc(int l, int c){
    L = l;
    C = c;
    T = new G*[L];
    for(int i=0; i<L; i++)
        T[i] = new G[C];
}

template<class G>
void DamierDynG<G>::Print(){
    cout << endl;
    for(int i=0; i<L; i++) {
        cout << endl;
        for(int j=0; j<C; j++)
            cout << T[i][j] << " ";
    }
}

template<class G>
void DamierDynG<G>::Init(G value){
    for(int i=0; i<L; i++)
        for(int j=0; j<C; j++)
            T[i][j]=value;
}

template<class G>
void DamierDynG<G>::Set(int x, int y, G value) {
    T[x][y]=value;
}

template<class G>
void DamierDynG<G>::ReDim(int l, int c, G vd) {
    Free();
    Alloc(l, c);
    Init(vd);
}

template<class G>
DamierDynG<G>& DamierDynG<G>::operator= (const DamierDynG<G> &D){
    if ( this != &D ) { // protection autoréférence
        Free();
        Alloc(D.L, D.C);
        for(int i=0; i<L; i++)
            for(int j=0; j<C; j++)
                T[i][j] = D.T[i][j];
    }
    return *this;
}

template<class G>
bool DamierDynG<G>::sameDimensions (const DamierDynG<G> &D) {
    if ((L == D.L) && (C == D.C))
        return true;
    return false;
}

template<class G>
DamierDynG<G>& DamierDynG<G>::operator+= (G c)
{
    for(int i=0; i<L; i++)
        for(int j=0; j<C; j++)
            T[i][j] += c;
    return *this;
}

template<class G>
DamierDynG<G>& DamierDynG<G>::operator+= (const DamierDynG<G> &D)
{
    if (!sameDimensions(D)){
        cerr << endl << __PRETTY_FUNCTION__ << " - Dimensions différentes !\n";
        exit(1);
    }
    for(int i=0; i<L; i++)
        for(int j=0; j<C; j++)
            T[i][j] += D.T[i][j];
    return *this;
}

template<class G>
DamierDynG<G> DamierDynG<G>::operator+ (const DamierDynG<G> &D1)
{
    if (!sameDimensions(D1)){
        cerr << endl << __PRETTY_FUNCTION__ << " - Dimensions différentes !\n";
        exit(1);
    }

    DamierDynG<G> D(D1.L, D1.C);
    for(int i=0; i<L; i++)
        for(int j=0; j<C; j++)
            D.T[i][j] = T[i][j]+D1.T[i][j];

    return D;
}

template<>
DamierDynG<bool> DamierDynG<bool>::operator+ (const DamierDynG<bool> &D1)
{
    cout << endl << "Opérateur + spécifique au type bool" << endl;
    if (!sameDimensions(D1)){
        cerr << endl << __PRETTY_FUNCTION__ << " - Dimensions différentes !\n";
        exit(1);
    }

    DamierDynG<bool> D(D1.L, D1.C);
    for(int i=0; i<L; i++)
        for(int j=0; j<C; j++)
            D.T[i][j] = T[i][j]||D1.T[i][j];

    return D;
}


template <class G>
ostream& operator<< (ostream& sortie, DamierDynG<G>& V)
{
    for(int i=0; i<V.L; i++) {
        for(int j=0; j<V.C; j++)
            sortie << V.T[i][j] << " ";
        sortie << endl;
    }

    return sortie;
}


#endif // DAMIERDYNG_H
